from typing import Any, Callable, List, Tuple, Type, TypeVar, Union, Dict
from functools import partial
from numpy import ndarray
from pathlib import Path
from torch import Tensor
import numpy as np
import torch
import yaml

# export list
__all__ = [
    # types
    "OneOrPair",
    "TensorOrArray",
    # pytorch devices
    "get_device",
    "set_device",
    "auto_device",
    # generic utilities
    "DotDict",
    "Config",
    "get_np_or_torch",
    "pack_many",
    "repeat_test",
    "execute_with_probability",
    # tensor/array operations
    "to_numpy",
    "split_complex",
    "pack_complex",
    "phase",
    # math utilities
    "factorize",
]

#  types

""" 
Generic variable
"""
T = TypeVar("T")

"""
Single or Pair of values of a certain type.
"""
OneOrPair = Union[T, Tuple[T, T]]

"""
Can be a torch Tensor or numpy ndarray.
"""
TensorOrArray = Union[Tensor, ndarray]


#  handling pytorch devices
def get_device() -> str:
    """
    Gets the first CUDA device available or CPU
    if no CUDA device is available.

    Returns
    -------
    str
        Device id
    """
    return "cuda" if torch.cuda.is_available() else "cpu"


def set_device(device: str, dtype: str = "Float") -> None:
    """
    Sets the default pytorch tensor
    to 'torch.{device}.FloatTensor'

    if device == "auto" it's inferred from get_device()

    Parameters
    ----------
    device : str
        Name of the device or "auto"
    dtype : str, optional
        Type of the tensor, by default "Float"
    """
    if device == "auto":
        device = get_device()
    if device == "cpu":
        torch.set_default_tensor_type(f"torch.{dtype}Tensor")
    else:
        torch.set_default_tensor_type(f"torch.{device}.{dtype}Tensor")


def auto_device(dtype: str = "Float") -> Callable:
    def _auto_device(f: Callable) -> Callable:
        """
        Decorator to set the pytorch device to auto.

        Parameters
        ----------
        f : Callable
            Function to decorate
        dtype : str, optional
            Type of the tensor, by default "Float"

        Returns
        -------
        Callable
            Decorated function
        """
        set_device("auto", dtype)
        return f

    return _auto_device


auto_device = partial(auto_device)

# generic utilities


class DotDict(dict):
    """
    A dict that allows dot notation
    for accessing to the elements
    """

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def copy(self) -> "DotDict":
        return DotDict(super().copy())


class Config:
    def __init__(self, config_path: Path) -> None:
        """
        A class to handle yaml configuration files.
        Inside a configuration many sections can be defined,
        with parameters within them.

        Parameters
        ----------
        config_path : Path
            Path to the yaml configuration file
        """
        # loading the configuration
        err_msg = "only YAML configurations are supported"
        assert config_path.suffix == ".yml", err_msg
        with open(config_path, "r") as f:
            self.config = yaml.safe_load(f)

    def get(
        self,
        section: str,
        parameter: str,
        _type: Type = str,
        default: Any = None,
    ) -> Any:
        """
        Gets a parameter from the configuration.

        Parameters
        ----------
        section : str
            Configuration section
        parameter : str
            Name of the parameter
        _type : Type, optional
            Type of the parameter, by default str
        default : Any, optional
            Default if the parameter does not exist, by default None

        Returns
        -------
        Fetched parameter
        """
        cfg = self.config

        try:
            # getting the section
            sec = cfg[section]
            # getting the parameter
            param = _type(sec[parameter])
        except KeyError:
            return default

        return param


def get_np_or_torch(x: TensorOrArray):
    """
    Returns numpy or torch modules depending on the input

    Parameters
    ----------
    x : TensorOrArray
        Input

    Returns
    -------
    Module
        numpy or torch
    """
    if isinstance(x, Tensor):
        return torch
    else:
        torch.iscomplex = torch.is_complex
        torch.hanning = torch.hann_window
        np.cat = np.concatenate
        return np


def pack_many(*xss: List[List]) -> List[Tuple]:
    """
    Packs many Lists in one.

    Parameters
    ----------
    xss : List[List]

    Returns
    -------
    List[Tuple]
        Packed list
    """
    return list(zip(*xss))


# fmt: off
def repeat_test(times: int): 
    """
    Decorator to repeat a test n times.

    Parameters
    ----------
    times : int
        Number of repetitions
    """
    def repeatHelper(f):
        def callHelper(*args):
            for _ in range(0, times):
                f(*args)
        return callHelper
    return repeatHelper 
# fmt: on


def execute_with_probability(prob: float) -> bool:
    """
    Returns True of false based on the
    random variable prob

    Parameters
    ----------
    prob : float
        Probability to execute, from 0 (never happen)
        to 1 (always happen)

    Returns
    -------
    bool
        True prob % of the times
    """
    return np.random.uniform() > prob


# tensor/array operations


def to_numpy(x: Tensor) -> np.ndarray:
    """
    Converts a Tensor into a numpy array.

    Parameters
    ----------
    x : Tensor
        Original tensor

    Returns
    -------
    np.ndarray
        Converted np array
    """
    return x.cpu().detach().numpy()


def split_complex(x: TensorOrArray) -> TensorOrArray:
    """
    Splits a complex Tensor in a float
    Tensor with the double of the channels.

    Parameters
    ----------
    x : TensorOrArray
        Complex input of shape (B, C, ...)

    Returns
    -------
    TensorOrArray
        Float output of shape (B, 2*C, ...)
    """
    module = get_np_or_torch(x)
    x = module.cat((x.real, x.imag), 1)
    return x


def pack_complex(x: TensorOrArray) -> TensorOrArray:
    """
    Merges a 2*C channels float TensorOrArray in a complex TensorOrArray.

    Parameters
    ----------
    x : TensorOrArray
        Float input of shape (B, 2*C, ...)

    Returns
    -------
    TensorOrArray
        Complex output of shape (B, C, ...)
    """
    c = x.shape[1]
    x = x[:, : c // 2] + 1j * x[:, c // 2 :]
    return x


def phase(x: TensorOrArray) -> TensorOrArray:
    """
    Computes the phase of a complex signal x
    as in: phase(x) = exp{j*angle(x)}

    Parameters
    ----------
    x : TensorOrArray
        Input complex signal

    Returns
    -------
    TensorOrArray
        Elementwise phase
    """
    module = get_np_or_torch(x)
    y = module.exp(1j * module.angle(x))
    return y


# math utilities


def factorize(n: int) -> List[int]:
    """
    Factorize an integer number.
    Implementation based on
    https://stackoverflow.com/questions/16007204/factorizing-a-number-in-python

    Parameters
    ----------
    n : int
        Number to factorize

    Returns
    -------
    List[int]
        List of factors in increasing order
    """

    def get_factor(n):
        j = 2
        while n > 1:
            for i in range(j, int(np.sqrt(n + 0.05)) + 1):
                if n % i == 0:
                    n //= i
                    j = i
                    yield i
                    break
            else:
                if n > 1:
                    yield int(n)
                    break

    factors = list(get_factor(n))
    factors = sorted(factors)
    return factors



from torch.utils.data import Sampler, Dataset, DataLoader, BatchSampler
from typing import List, Optional, Tuple
from pathlib import Path
from torch import Tensor
from tqdm import tqdm
import numpy as np
import itertools
import torch
import h5py


from torch_utilities.common import to_numpy, split_complex

# export list
__all__ = [
    "HDF5Dataset",
    "get_hdf5_dataloader",
    "HDF5OnlineDataset",
    "get_dataset_statistics",
    "plot_dataset_statistics",
]


class WeakShufflingSampler(Sampler):
    def __init__(self, dataset: Dataset, batch_size: int):
        """
        Sampler that implements weak-shuffling.

        E.g. if dataset is [1,2,3,4] with batch_size=2,
             then first batch, [[1,2], [3,4]] then
             shuffle batches -> [[3,4],[1,2]]

        Referenece:
        https://towardsdatascience.com/reading-h5-files-faster-with-pytorch-datasets-3ff86938cc

        Parameters
        ----------
        dataset : Dataset
            Source dataset
        batch_size : int
            Size of the batches
        """
        self.batch_size = batch_size
        self.dataset_length = len(dataset)
        self.n_batches = int(np.ceil((self.dataset_length / self.batch_size)))
        self.batch_ids = torch.randperm(self.n_batches)

    def __len__(self):
        return self.batch_size

    def __iter__(self):
        datalen = self.dataset_length
        for id in self.batch_ids:
            start = id * self.batch_size
            end = (id + 1) * self.batch_size
            end = end if end < datalen else datalen
            selection = torch.arange(start, end)
            for index in selection:
                yield int(index)


class HDF5Dataset(Dataset):
    def __init__(
        self,
        dataset_path: Path,
        data_layout: List[str],
    ) -> None:
        """
        Dataset supporting HDF5 format.

        The following constraints should be satisfied
        by the HDF5 file for a correct behaviour:
        - Multiple single-level groups (e.g. /group1, /group2, ...)
        - The batch size (first dimension) of each dataset is the same
        - Each group has len(data_layout) dataset inside

        Parameters
        ----------
        dataset_path : Path
            Path to the .hdf5 file
        data_layout : List[str]
            List describing the layout of the data
            inside each group. For an input-label1-label2
            dataset the list would be ["input", "label1", "label2"]
        """
        super().__init__()
        self.dataset_path = dataset_path.absolute()
        self.data_layout = data_layout
        self.dataset_file = h5py.File(self.dataset_path, "r")
        self.groups = list(self.dataset_file.keys())
        self.group_batch_len = self.dataset_file[self.groups[0]][
            self.data_layout[0]
        ].shape[0]
        self.data_layout = data_layout
        self._cache = None
        self._cache_idx = None

    def __len__(self):
        return len(self.groups) * self.group_batch_len

    def __getitem__(self, idx) -> List[Tensor]:
        # error handling
        err_msg = None
        if not isinstance(idx, list):
            err_msg = "HDF5Dataset must be sampled by a BatchLoader"
        elif len(idx) > self.group_batch_len:
            err_msg = (
                "Reduce the batch size to be less than the batch size of the groups"
            )
        elif self.group_batch_len % len(idx) != 0:
            err_msg = (
                "Modify the batch size to be a divider of the HDF5 group batch size"
            )
        if err_msg is not None:
            raise RuntimeError(err_msg)

        # cache update
        if not self._in_cache(idx[0]):
            self._update_cache(idx[0])

        # using the cache
        gbl = self.group_batch_len
        a, b = idx[0] % gbl, (idx[-1] % gbl) + 1
        data = [x[a:b] for x in self._cache]

        return data

    def _update_cache(self, idx: int) -> None:
        """
        Caches a group in memory.

        Parameters
        ----------
        idx : int
            Index of an element of the group
        """
        # getting the starting index of a group
        self._cache_idx = idx

        # updating the cache
        del self._cache
        g_idx = idx // self.group_batch_len
        g = self.dataset_file[self.groups[g_idx]]
        cast = lambda x: torch.from_numpy(np.array(x)).to("cpu")
        data = [cast(g[k]) for k in self.data_layout]
        self._cache = data

    def _in_cache(self, idx: int) -> bool:
        """
        Checks if an index is inside the cache.

        Parameters
        ----------
        idx : int
            Target index

        Returns
        -------
        bool
            True if the element is inside the cache
        """
        c_idx = self._cache_idx
        flag = (c_idx is not None) and (
            idx >= c_idx and idx < (c_idx + self.group_batch_len)
        )
        return flag


def get_hdf5_dataloader(
    dataset_path: Path,
    data_layout: List[str],
    batch_size: int = None,
    **dataloader_kwargs: dict,
) -> DataLoader:
    """
    Create a dataloader binded to a HDF5Dataset.

    Parameters
    ----------
    dataset : Path
        Path to the HDF5 Dataset
    data_layout : List[str]
            List describing the layout of the data
            inside each group. For an input-label1-label2
            dataset the list would be ["input", "label1", "label2"]
    batch_size : int, optional
        Batch size of the dataloader, by default HDF5Dataset.group_batch_len
    dataloader_kwargs : dict, optional
        DataLoader arguments, by default {}
    """
    if batch_size is None:
        batch_size = dataset.group_batch_len

    if dataloader_kwargs is None:
        dataloader_kwargs = {}

    dataset = HDF5Dataset(dataset_path, data_layout)

    sampler = BatchSampler(
        sampler=WeakShufflingSampler(
            dataset=dataset,
            batch_size=batch_size,
        ),
        batch_size=batch_size,
        drop_last=False,
    )

    dataloader = DataLoader(
        dataset=dataset,
        sampler=sampler,
        **dataloader_kwargs,
    )

    return dataloader


class HDF5OnlineDataset(Dataset):
    def __init__(
        self,
        dataset_paths: List[Path],
        data_layouts: List[List[str]],
        batch_size: int = 16,
        total_items: int = 1024,
        overfit_mode: bool = False,
    ) -> None:
        """
        Dataset based on online generation.
        Data can be combined after retrieving it
        from source HDF5Datasets.

        Parameters
        ----------
        dataset_paths : List[Path]
            List of paths to the HDF5Datasets h5
        data_layouts : List[List[str]]
            Data layouts for all the HDF5Datasets
        batch_size : int, optional
            Batch size of each raw data, by default 16
        total_items : int, optional
            Total item in an epoch, by default 1024
        overfit_mode : bool, optional
            If True, always returns the first batch at every iteration
        """
        super(HDF5OnlineDataset).__init__()
        self.datasets_paths = dataset_paths
        self.data_layouts = data_layouts
        self.batch_size = batch_size
        self.total_items = total_items
        self.overfit_mode = overfit_mode

        # error handling
        err_msg = f"total_items ({total_items}) must be divisible by batch_size ({batch_size})"
        assert (self.total_items % self.batch_size) == 0, err_msg

        # static datasets
        self.source_datasets = self._get_datasets()

        # handling overfit mode
        self.overfit_cache = None

    def __len__(self) -> int:
        return self.total_items // self.batch_size

    def transform(self, raw_data: List[Tensor]) -> List[Tensor]:
        """
        Override this method to process the raw data.

        Parameters
        ----------
        raw_data : List[Tensor]
            Raw data from the HDF5Datasets

        Returns
        -------
        List[Tensor]
            Processed data
        """
        return raw_data

    def _get_rand_batch(self, dataset: HDF5Dataset) -> List[Tensor]:
        """
        Get a random batch from an HDF5Dataset

        Parameters
        ----------
        dataset : HDF5Dataset
            Target HDF5Dataset

        Returns
        -------
        List[Tensor]
            Batches from the HDF5Datasets
        """
        # getting start and end points
        ds_len = len(dataset)
        total_batches = ds_len // self.batch_size
        r_int = np.random.randint(0, total_batches)
        start = r_int * self.batch_size
        end = (r_int + 1) * self.batch_size
        indices = list(range(start, end))

        # getting the random selection
        ds_outs = dataset[indices]
        return ds_outs

    def __getitem__(self, _) -> List[Tensor]:
        if self.overfit_cache is not None:
            # overfit cache enabled
            return self.overfit_cache

        outs = [self._get_rand_batch(ds) for ds in self.source_datasets]
        outs = itertools.chain.from_iterable(outs)
        outs = list(outs)
        outs = self.transform(outs)

        if self.overfit_mode:
            # caching for overfit mode
            self._store_in_cache(outs)

        return outs

    def _get_datasets(self) -> List[HDF5Dataset]:
        """
        Gets the static datasets.

        Returns
        -------
        List[HDF5Dataset]
            List of static datasets
        """
        ds = [HDF5Dataset(p, l) for p, l in zip(self.datasets_paths, self.data_layouts)]
        return ds

    def _store_in_cache(self, data: List[Tensor]) -> None:
        """
        Stores the data in the overfit cache.

        Parameters
        ----------
        data : List[Tensor]
            Data to store in the cache
        """
        self.overfit_cache = data


def get_dataset_statistics(
    dataloader: DataLoader,
    iterations: int,
    n_hist_bins: int = 100,
    verbose: bool = False,
) -> List[Tuple[np.ndarray, np.ndarray, float, float, float]]:
    """
    Calculate the statistics for the data obtained
    from a dataset.

    Parameters
    ----------
    dataloader : DataLoader
        DataLoader to iterate
    iterations : int
        Number of iterations
    n_hist_bins : int, optional
        Number of histogram bins
    verbose : bool, optional
        If True shows a progress bar for the iterations,
        by default False

    Returns
    -------
    List[Tuple[np.ndarray, np.ndarray, float, float, float, float]]
        List of statistics for each dataset output
        each tuple contains
        (hist_vals, hist_bins, min, max, mean, var)
    """
    ds = iter(dataloader)
    _range = tqdm(range(iterations)) if verbose else range(iterations)
    outs = [ds.__next__() for i in _range]
    n_items = len(outs[0])

    stats = []
    for i in range(n_items):
        x = torch.stack([o[i] for o in outs])
        x = x.flatten()
        x = x.float()
        x = split_complex(x) if torch.is_complex(x) else x
        hist_vals, hist_bins = [
            to_numpy(y) for y in x.histogram(n_hist_bins, density=True)
        ]
        hist_vals /= n_hist_bins
        _min = x.min().item()
        _max = x.max().item()
        _mean = x.mean().item()
        _var = x.var().item()
        stats.append((hist_vals, hist_bins, _min, _max, _mean, _var))

    return stats


def plot_dataset_statistics(
    dataloader: DataLoader,
    iterations: int,
    n_hist_bins: int = 100,
    labels: Optional[List[str]] = None,
    save_path: Optional[Path] = None,
    figsize: Tuple[int, int] = (8, 6),
    verbose: bool = False,
) -> List[Tuple[np.ndarray, np.ndarray, float, float, float, float]]:
    """
    Plots the statistics for the data obtained
    from a dataset.

    Parameters
    ----------
    dataloader : DataLoader
        DataLoader to iterate
    iterations : int
        Number of iterations
    n_hist_bins : int, optional
        Number of histogram bins
    labels : Optional[List[str]]
        Names of the dataset outputs
    save_path : Optional[Path]
        If specified saves the plots in the specified folder
    figsize : Tuple[int, int], optional
        Size of the plot, by default (8, 6)
    verbose : bool, optional
        If True shows a progress bar for the iterations,
        by default False


    Returns
    -------
    List[Tuple[np.ndarray, np.ndarray, float, float, float]]
        List of statistics for each dataset output
        each tuple contains
        (hist_vals, hist_bins, min, max, mean, var)
    """
    import matplotlib.pyplot as plt

    all_stats = get_dataset_statistics(
        dataloader,
        iterations,
        n_hist_bins,
        verbose,
    )

    # default labels
    if labels is None:
        labels = [f"feature_{i}" for i in range(len(all_stats))]

    # optional save path
    if save_path is not None:
        save_path = Path(save_path)

    for stats, name in zip(all_stats, labels):
        hist_vals, hist_bins, min, max, mean, var = stats
        bar_width = np.abs(hist_bins[1] - hist_bins[0])
        plt.figure(figsize=figsize)
        plt.title(f"{name} distribution")
        plt.bar(hist_bins[:-1], hist_vals, width=bar_width)
        plt.vlines(x=mean, ymin=0, ymax=1, colors="red", label="mean")
        plt.ylim([0, hist_vals.max()])
        plt.legend()
        plt.grid()
        if save_path is not None:
            plt.savefig(save_path / f"{name}.png")
        else:
            plt.show()
        plt.close()

    return all_stats



from itertools import product
from pathlib import Path
from torch import Tensor
from typing import List
import numpy as np
import unittest
import torch

from tests.generate_test_data import get_test_data_dir
import torch_utilities as tu


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("cpu")
    torch.set_grad_enabled(False)


class TestHDF5DataLoader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.data_layout = ["x", "y_true"]
        self.hdf5_path = get_test_data_dir() / "dataset.hdf5"
        self.dataset = tu.HDF5Dataset(self.hdf5_path, self.data_layout)

    def dummy_input(self, g_idx):
        return torch.ones((16, 8)) * g_idx

    def test_weak_shuffling_len(self):
        batch_size = 16
        sampler = tu.data_loading.WeakShufflingSampler(self.dataset, batch_size)
        self.assertEqual(len(sampler), batch_size)

    def test_hdf5dataset_cache(self):
        group_len = 16

        self.assertEqual(self.dataset._cache, None)
        self.assertEqual(self.dataset._cache_idx, None)

        self.dataset[[0, 1]]
        self.assertTrue(torch.all(self.dataset._cache[0] == self.dummy_input(0)))
        self.assertTrue(torch.all(self.dataset._cache[1] == self.dummy_input(0)))
        self.assertEqual(self.dataset._cache_idx, 0)

        self.dataset[[group_len * 3, group_len * 3 + 1]]
        self.assertTrue(torch.all(self.dataset._cache[0] == self.dummy_input(3)))
        self.assertTrue(torch.all(self.dataset._cache[1] == self.dummy_input(3)))
        self.assertEqual(self.dataset._cache_idx, group_len * 3)

    def test_hdf5dataset_raise_idx_not_list(self):
        with self.assertRaises(RuntimeError):
            self.dataset[0]

    def test_hdf5dataset_raise_idx_len_too_long(self):
        with self.assertRaises(RuntimeError):
            self.dataset[list(range(17))]

    def test_hdf5dataset_raise_idx_len_not_divisor(self):
        with self.assertRaises(RuntimeError):
            self.dataset[[1, 2, 3]]

    def test_hdf5_dataloader_full_batch(self):
        dataloader = tu.get_hdf5_dataloader(self.hdf5_path, self.data_layout, 16)
        data = [x for x in dataloader]
        self.assertEqual(len(data), 10)

    def test_hdf5_dataloader_half_batch(self):
        dataloader = tu.get_hdf5_dataloader(self.hdf5_path, self.data_layout, 8)
        data = [x for x in dataloader]
        self.assertEqual(len(data), 20)

    def test_get_dataset_statistics(self):
        hist_bins = 10
        dataloader = tu.get_hdf5_dataloader(self.hdf5_path, self.data_layout, 16)
        iterations = len(dataloader)
        stats = tu.get_dataset_statistics(dataloader, iterations, hist_bins)
        hist_vals, hist_bins, min, max, mean, var = stats[0]
        self.assertEqual(len(stats), 2)
        self.assertAlmostEqual(np.sum(hist_vals), 1)


class HDF5OnlineDatasetTesting(tu.HDF5OnlineDataset):
    def __init__(
        self,
        dataset_paths: List[Path],
        data_layouts: List[List[str]],
        batch_size: int,
        total_items: int,
    ) -> None:
        super().__init__(dataset_paths, data_layouts, batch_size, total_items)

    def transform(self, raw_data: List[Tensor]) -> List[Tensor]:
        return raw_data


class TestHDF5OnlineDataset(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.data_layout = ["x", "y_true"]
        self.hdf5_path = get_test_data_dir() / "dataset.hdf5"

    def test_assert(self):
        batch_len = (16, 32)
        total_items = (77, 171)
        for b, t in product(batch_len, total_items):
            with self.subTest(b=b, t=t), self.assertRaises(AssertionError):
                HDF5OnlineDatasetTesting(
                    dataset_paths=[self.hdf5_path],
                    data_layouts=[self.data_layout],
                    batch_size=b,
                    total_items=t,
                )

    def test_len(self):
        batch_len = (8, 16)
        total_items = (16, 32)
        for b, t in product(batch_len, total_items):
            with self.subTest(b=b, t=t):
                ds = HDF5OnlineDatasetTesting(
                    dataset_paths=[self.hdf5_path],
                    data_layouts=[self.data_layout],
                    batch_size=b,
                    total_items=t,
                )
                ds_len = len(ds)
                self.assertEqual(ds_len, t // b)

    def test_get_datasets(self):
        ds = HDF5OnlineDatasetTesting(
            dataset_paths=[self.hdf5_path, self.hdf5_path],
            data_layouts=[self.data_layout, self.data_layout],
            batch_size=8,
            total_items=1024,
        )

        datasets = ds._get_datasets()
        for d in datasets:
            self.assertEqual(type(d), tu.HDF5Dataset)

    def test_get_rand_batch(self):
        batch_size = 8
        ds = HDF5OnlineDatasetTesting(
            dataset_paths=[self.hdf5_path],
            data_layouts=[self.data_layout],
            batch_size=batch_size,
            total_items=64,
        )

        datasets = ds._get_datasets()
        for d in datasets:
            raw_data = ds._get_rand_batch(d)
            for batch in raw_data:
                self.assertEqual(batch.shape[0], batch_size)

    def test_get_dataset_statistics(self):
        hist_bins = 10
        dataloader = tu.get_hdf5_dataloader(self.hdf5_path, self.data_layout, 16)
        iterations = len(dataloader)
        stats = tu.get_dataset_statistics(dataloader, iterations, hist_bins)
        hist_vals, hist_bins, min, max, mean, var = stats[0]
        self.assertEqual(len(stats), 2)
        self.assertAlmostEqual(np.sum(hist_vals), 1)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from torch_utilities.io import *
from torch_utilities.audio import *
from torch_utilities.common import *
from torch_utilities.pytorch import *
from torch_utilities.data_loading import *
import torch_utilities.augmentation
import torch_utilities.features
import torch_utilities.metrics
import torch_utilities.modules
import torch_utilities.losses



from torchaudio.functional import melscale_fbanks
from typing import Optional, List
import torch.nn.functional as F
from random import randrange
from torch import Tensor
import numpy as np
import torch


from torch_utilities.common import get_np_or_torch, TensorOrArray, to_numpy, get_device


# export list
__all__ = [
    "stft",
    "istft",
    "MelFilterbank",
    "MelInverseFilterbank",
    "db",
    "invert_db",
    "power",
    "energy",
    "rms",
    "snr",
    "fade_sides",
    "random_trim",
    "trim_silence",
    "interleave",
]


def stft(
    x: TensorOrArray,
    sample_rate: int = 16000,
    hopsize_ms: int = 10,
    window: str = "hann",
    win_len_ms: int = 20,
    win_oversamp: int = 2,
) -> TensorOrArray:
    """
    Calculates the STFT of a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)
    sample_rate : int, optional
        Sample rate of the signal, by default 16000
    hopsize_ms : int, optional
        STFT hopsize in ms, by default 10
    window : str, optional
        Torch window to use, by default "hann"
    win_len_ms : int, optional
        Window length in ms, by default 20 ms
    win_oversamp : int, optional
        Zero padding applied equal to the window length
        (1 equals to no zero pad), by default 2

    Returns
    -------
    TensorOrArray
        STFT of the input of shape (..., T', F')

    Raises
    ------
    AttributeError
        If the window chosen does not exist
    """
    return _stft_istft_core(
        True,
        x,
        sample_rate,
        hopsize_ms,
        window,
        win_len_ms,
        win_oversamp,
    )


def istft(
    x: TensorOrArray,
    sample_rate: int = 16000,
    hopsize_ms: int = 10,
    window: str = "hann",
    win_len_ms: int = 20,
    win_oversamp: int = 2,
) -> TensorOrArray:
    """
    Calculates the ISTFT of a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T, F)
    sample_rate : int, optional
        Sample rate of the signal, by default 16000
    hopsize_ms : int, optional
        STFT hopsize in ms, by default 10
    window : str, optional
        Torch window to use, by default "hann"
    win_len_ms : int, optional
        Window length in ms, by default 20 ms
    win_oversamp : int, optional
        Zero padding applied equal to the window length
        (1 equals to no zero pad), by default 2

    Returns
    -------
    TensorOrArray
        ISTFT of the input of shape (..., T')

    Raises
    ------
    AttributeError
        If the window chosen does not exist
    """
    return _stft_istft_core(
        False,
        x,
        sample_rate,
        hopsize_ms,
        window,
        win_len_ms,
        win_oversamp,
    )


def _stft_istft_core(
    is_stft: bool,
    x: TensorOrArray,
    sample_rate: int = 16000,
    hopsize_ms: int = 10,
    window: str = "hann",
    win_len_ms: int = 20,
    win_oversamp: int = 2,
) -> TensorOrArray:
    """
    Calculates the STFT/ISTFT of a signal.

    Parameters
    ----------
    is_stft : bool
        Selects between STFT and ISTFT
    x : TensorOrArray
        Input signal
    sample_rate : int, optional
        Sample rate of the signal, by default 16000
    hopsize_ms : int, optional
        STFT hopsize in ms, by default 10
    window : str, optional
        Torch window to use, by default "hann"
    win_len_ms : int, optional
        Window length in ms, by default 20 ms
    win_oversamp : int, optional
        Zero padding applied equal to the window length
        (1 equals to no zero pad), by default 2

    Returns
    -------
    TensorOrArray
        STFT of the input

    Raises
    ------
    AttributeError
        If the window chosen does not exist
    """
    # converting to Tensor
    in_type = type(x)
    if in_type == np.ndarray:
        x = torch.from_numpy(x)

    # getting the window function
    try:
        window += "_window"
        win_fun = getattr(torch, window)
    except AttributeError:
        allowed_win = [
            w + "_window" for w in ["hann", "hamming", "bartlett", "blackman", "kaiser"]
        ]
        err_msg = "choose a window between:\n" + ", ".join(allowed_win)
        raise AttributeError(err_msg)

    # parameters of the STFT
    _ms_to_samples = lambda x: int(x * sample_rate / 1000)
    win_len = _ms_to_samples(win_len_ms)
    hopsize = _ms_to_samples(hopsize_ms)
    n_fft = int(win_len * win_oversamp)
    _window = torch.zeros(n_fft, device=x.device)
    _window[:win_len] = win_fun(win_len, device=x.device)

    # STFT/ISTFT dependent code
    _transpose = lambda x: x.transpose(-1, -2)
    if is_stft:
        transform = torch.stft
        # compensating for oversampling and center==True
        pad_ovr = n_fft - win_len
        pad_ctr = win_len // 2
        x = F.pad(x, (pad_ctr, pad_ovr))
    else:
        transform = torch.istft
        x = _transpose(x)

    try:
        y = transform(
            x,
            n_fft=n_fft,
            hop_length=hopsize,
            window=_window,
            return_complex=is_stft,
            center=True,
        )
    except RuntimeError as e:
        # TODO: pytorch bug
        # for some reason, few configurations do not pass the NOLA check even if
        # they should pass it. This seems to be a bug related to PyTorch
        err_msg = "The configuration you provided does not satisfies PyTorch NOLA check"
        raise RuntimeError(str(e) + "\n" + err_msg)

    if is_stft:
        # reshaping
        y = _transpose(y)
        # compensating for center==True
        y = y[:, 1:]
    else:
        # compensating for ola ratio
        # it's still not clear where this shift comes from
        ola_shift = (win_len // hopsize - 2) * (hopsize // 2)
        y = y[:, ola_shift:]

    if in_type == np.ndarray:
        # converting to numpy
        y = to_numpy(y)

    return y


class MelFilterbank:
    def __init__(
        self,
        sample_rate: int,
        n_freqs: int,
        n_mels: int,
        device: Optional[torch.device] = None,
    ) -> None:
        """
        Apply Mel filterbank to the input batch.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        n_freqs : int
            Stft frequencies
        n_mels : int
            Number of mel frequencies
        device : Optional[torch.device], optional
            Device for the filterbank matrix, by default None
        """
        self.sample_rate = sample_rate
        self.n_freqs = n_freqs
        self.n_mels = n_mels
        self.device = device or get_device()
        self.filterbank = self._get_filterbank()

    def _get_filterbank(self) -> Tensor:
        """
        Gets the mel filterbank

        Returns
        -------
        Tensor
            Mel filterbank matrix
        """
        filterbank = melscale_fbanks(
            n_freqs=self.n_freqs,
            n_mels=self.n_mels,
            f_min=30,
            f_max=self.sample_rate // 2,
            sample_rate=self.sample_rate,
        )
        filterbank = filterbank.to(self.device)

        return filterbank

    def __call__(self, x: TensorOrArray) -> Tensor:
        """
        Parameters
        ----------
        x : TensorOrArray
            Signal of shape (B, C, T, n_freq)

        Returns
        -------
        Tensor
            STFT of shape (B, C, T, n_mel)
        """
        is_np = get_np_or_torch(x) == np
        if is_np:
            x = torch.from_numpy(x).to(self.device)

        # handling dtypes
        x = x.abs()
        self.filterbank = self.filterbank.to(x.dtype)

        y = x @ self.filterbank

        if is_np:
            y = to_numpy(y)
        return y


class MelInverseFilterbank:
    def __init__(
        self,
        sample_rate: int,
        n_freqs: int,
        n_mels: int,
        device: Optional[torch.device] = None,
    ) -> None:
        """
        Apply inverse Mel filterbank to the input batch,
        to get back a spectrogram.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        n_freqs : int
            Stft frequencies
        n_mels : int
            Number of mel frequencies
        device : Optional[torch.device], optional
            Device for the filterbank matrix, by default None
        """
        self.sample_rate = sample_rate
        self.n_freqs = n_freqs
        self.n_mels = n_mels
        self.device = device or get_device()
        self.filterbank = self._get_filterbank()

    def _get_filterbank(self) -> Tensor:
        """
        Gets the mel filterbank

        Returns
        -------
        Tensor
            Mel filterbank matrix
        """
        filterbank = melscale_fbanks(
            n_freqs=self.n_freqs,
            n_mels=self.n_mels,
            f_min=30,
            f_max=self.sample_rate // 2,
            sample_rate=self.sample_rate,
        )
        # pseudo-inverse is used to approximate
        # the inverse transform
        filterbank = torch.linalg.pinv(filterbank)
        filterbank = filterbank.to(self.device)

        return filterbank

    def __call__(self, x: TensorOrArray) -> TensorOrArray:
        """
        Parameters
        ----------
        x : TensorOrArray
            Signal of shape (B, C, T, n_freq)

        Returns
        -------
        TensorOrArray
            STFT of shape (B, C, T, n_mel)
        """
        is_np = get_np_or_torch(x) == np
        if is_np:
            x = torch.from_numpy(x).to(self.device)

        # handling dtypes
        x = x.abs()
        self.filterbank = self.filterbank.to(x.dtype)

        y = x @ self.filterbank

        if is_np:
            y = to_numpy(y)
        return y


def db(x: TensorOrArray, eps: float = 1e-12) -> TensorOrArray:
    """
    Converts linear to dB

    Parameters
    ----------
    x : TensorOrArray
        Input signal amplitude
    eps : float
        Number summed to x before applying the logarithm,
        by default 1e-12

    Returns
    -------
    float
        Input in dB
    """
    module = get_np_or_torch(x)
    return 20 * module.log10(x + eps)


def invert_db(x: TensorOrArray, eps: float = 1e-12) -> TensorOrArray:
    """
    Converts dB to linear

    Parameters
    ----------
    x : TensorOrArray
        Input signal amplitude in dB

    Returns
    -------
    float
        Input inverting dB
    """
    return 10 ** (x / 20) - eps


def power(x: TensorOrArray) -> float:
    """
    Power of a signal, calculated for each channel.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        Power of the signal of shape (...) (len(x) - 1)
    """
    module = get_np_or_torch(x)
    _power = module.einsum("...t,...t->...", x, x.conj())
    return _power


def energy(x: TensorOrArray) -> float:
    """
    Energy of a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        Energy of the signal of shape (...) (len(x) - 1)
    """
    samples = x.shape[-1]
    return power(x) / samples


def rms(x: TensorOrArray) -> float:
    """
    RMS of a signal, calculated for each channel.

    Parameters
    ----------
    x : TensorOrArray
        Input signal of shape (..., T)

    Returns
    -------
    float
        RMS of the signal of shape (...) (len(x) - 1)
    """
    module = get_np_or_torch(x)
    return module.sqrt(energy(x))


def snr(x: TensorOrArray, noise: TensorOrArray) -> float:
    """
    Signal to Noise Ratio (SNR) ratio in dB,
    calculated considering the RMS.

    Parameters
    ----------
    x : TensorOrArray
        Signal of interest
    noise : TensorOrArray
        Interference

    Returns
    -------
    float
        SNR in db
    """
    err_msg0 = "snr supports only 1D and 2D signals"
    assert len(x.shape) in [1, 2], err_msg0
    assert len(noise.shape) in [1, 2], err_msg0
    err_msg1 = "x and noise should be of the same type"
    assert type(x) == type(noise), err_msg1

    module = get_np_or_torch(x)
    channel_mean = lambda x: module.mean(x, -2) if len(x.shape) == 1 else x
    a = channel_mean(db(rms(x)))
    b = channel_mean(db(rms(noise)))
    snr = a - b
    return snr


def _win_to_sides(
    x: TensorOrArray,
    win: TensorOrArray,
    fade_len: int,
    direction: str,
) -> TensorOrArray:
    """
    Handler used to apply a window over the sides of
    a signal.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., C, T)
    win : TensorOrArray
        Window
    fade_len : TensorOrArray
        Length of each fade in samples
    direction : str
        Indicates the sides in which the fade is applied,
        one between "left", "right" or "both"

    Returns
    -------
    TensorOrArray
        Faded output
    """
    # error handling
    err_msg = 'direction must be one between "left", "right" or "both"'
    assert direction in ["left", "right", "both"], err_msg

    if direction in ["left", "both"]:
        x[..., :fade_len] *= win[:fade_len]
    if direction in ["right", "both"]:
        x[..., -fade_len:] *= win[-fade_len:]
    return x


def fade_sides(
    x: TensorOrArray, fade_len: int = 100, direction: str = "both"
) -> TensorOrArray:
    """
    Apply an half of an Hanning window to the
    sides of the input, in order to obtain a fade in/out.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., C, T)
    fade_len : int, optional
        Length of the fade in samples, by default 10.
        The length of the window is 2 * fade_len + 1.
    direction : str
        Indicates the sides in which the fade is applied,
        one between "left", "right" or "both", by default "both"

    Returns
    -------
    TensorOrArray
        Faded output
    """
    module = get_np_or_torch(x)
    win = module.hanning(2 * fade_len + 1)
    if module == np:
        y = x.copy()
    else:
        win = win.to(get_device())
        win[-1] = 0
        y = x.detach().clone()
    y = _win_to_sides(y, win, fade_len, direction)

    return y


def random_trim(
    x: TensorOrArray,
    sample_rate: int,
    duration: float = 3,
) -> TensorOrArray:
    """
    Extracts a random temporal selection from the input.

    Parameters
    ----------
    x : TensorOrArray
        Input of shape (..., T)
    sample_rate : int
        Sample rate in Hz
    duration : float, optional
        Duration of the selection, by default 3 s

    Returns
    -------
    TensorOrArray
        Random temporal selection of the input
    """
    module = get_np_or_torch(x)
    x_len = x.shape[-1]
    duration_samples = int(duration * sample_rate)
    selection_start_max = x_len - duration_samples

    if selection_start_max <= 0:
        # if the input is longer than the selection
        # just zero pad the input
        y = module.zeros((*x.shape[:-1], duration_samples))
        y[..., :x_len] = x
    else:
        # applying selection
        start = randrange(0, selection_start_max)
        end = start + duration_samples
        y = x[..., start:end]
    return y


def trim_silence(
    x: TensorOrArray,
    threshold_db: float = -35,
    margin: int = 0,
) -> TensorOrArray:
    """
    Trims the silences at the beginning and end of a sample.

    Parameters
    ----------
    x : TensorOrArray
        Input sample of shape (T,)
    threshold_db : float, optional
        Relative to x.max() to detect the silences, by default -35 dB
    margin : int, optional
        Samples kept at both sides after the trimming, by default 0

    Returns
    -------
    TensorOrArray
        Trimmed ouput of shape (T',)
    """
    module = get_np_or_torch(x)

    # finding the start and end points
    threshold = invert_db(threshold_db)
    threshold *= module.abs(x).max()
    thr = module.zeros_like(x, dtype=int)
    thr[module.abs(x) > threshold] = 1
    thr = thr if module == np else thr.cpu().detach().numpy()
    thr = thr.tolist()

    try:
        start = thr.index(1)
        end = len(thr) - thr[::-1].index(1)

        # trimming the silences
        x_start = int(np.clip(start - margin, 0, None))
        x_end = end + margin
        y = x[..., x_start:x_end]
        return y
    except ValueError:
        # no value found
        return x


def interleave(*xs: List[TensorOrArray]) -> Tensor:
    """
    Interleaves many input tensors in one over the last dimension.

    Parameters
    ----------
    xs : List[TensorOrArray]
        Input signals of the same shape

    Returns
    -------
    Tensor
        Interleaved signal, the shape is the same of the inputs except
        for the interleaved dimension of lenght D that will be len(xs)*D
    """
    mod = get_np_or_torch(xs[0])
    stride = len(xs)
    new_shape = list(xs[0].shape)
    new_shape[-1] *= stride
    y = mod.zeros(new_shape)

    for i, x in enumerate(xs):
        y[..., i::stride] = x

    return y



from contextlib import suppress
from itertools import product
from torch import Tensor
import numpy as np
import unittest
import torch


import torch_utilities as tu
from tests.generate_test_data import get_test_data_dir


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestSTFT(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        module = (np, torch)
        channels = (1, 4)
        sample_rate = (8000, 16000, 48000)
        hopsize_ms = (4, 8)
        ola_ratio = (2, 3, 4)
        win_oversamp = (1, 2)
        self.params = product(
            module,
            channels,
            sample_rate,
            hopsize_ms,
            ola_ratio,
            win_oversamp,
        )

    def test_stft(self):
        for p in self.params:
            (
                mod,
                channels,
                sample_rate,
                hopsize_ms,
                ola_ratio,
                win_oversamp,
            ) = p
            with self.subTest(p=p):
                x = mod.ones((channels, sample_rate * 1))
                win_len_ms = hopsize_ms * ola_ratio
                y = tu.stft(
                    x, sample_rate, hopsize_ms, "hamming", win_len_ms, win_oversamp
                )
                bins = int(sample_rate * win_len_ms * win_oversamp / 2000 + 1)
                self.assertEqual(y.shape[-1], bins)

    def test_istft(self):
        for p in self.params:
            (
                mod,
                channels,
                sample_rate,
                hopsize_ms,
                ola_ratio,
                win_oversamp,
            ) = p
            with self.subTest(p=p):
                win_len_ms = hopsize_ms * ola_ratio
                bins = int(sample_rate * win_len_ms * win_oversamp / 2000 + 1)
                x = mod.ones((channels, int(sample_rate * 0.05), bins)) + 0j
                y = tu.istft(
                    x, sample_rate, hopsize_ms, "hann", win_len_ms, win_oversamp
                )

    def test_inversion(self):
        for p in self.params:
            (
                mod,
                channels,
                sample_rate,
                hopsize_ms,
                ola_ratio,
                win_oversamp,
            ) = p
            with self.subTest(p=p):
                if mod == np:
                    x = np.random.normal(size=(channels, sample_rate * 1))
                else:
                    x = torch.randn((channels, sample_rate * 1))

                win_len_ms = hopsize_ms * ola_ratio
                y = tu.stft(
                    x, sample_rate, hopsize_ms, "hann", win_len_ms, win_oversamp
                )
                x_hat = tu.istft(
                    y, sample_rate, hopsize_ms, "hann", win_len_ms, win_oversamp
                )
                x = x[0, : x_hat.shape[1]]
                x_hat = x_hat[0, : x.shape[0]]
                err = mod.abs(x - x_hat).max()
                with suppress(Exception):
                    err = err.item()
                self.assertAlmostEqual(err, 0, places=5)


class TestMelFilterbank(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        module = (np, torch)
        channels = (1, 4)
        sample_rate = (8000, 16000, 24000, 48000)
        n_freq = (160, 320, 640, 1280)
        n_mel = (8, 16, 32)

        self.params = product(
            module,
            channels,
            sample_rate,
            n_freq,
            n_mel,
        )

    def test_to_mel(self):
        for module, channels, sample_rate, n_freq, n_mel in self.params:
            with self.subTest(
                module=module,
                channels=channels,
                sample_rate=sample_rate,
                n_freq=n_freq,
                n_mel=n_mel,
            ):
                filterbank = tu.MelFilterbank(sample_rate, n_freq, n_mel)
                x = np.ones(
                    (1, channels, int(sample_rate * 0.1), n_freq), dtype=complex
                )
                x = x if module == np else torch.from_numpy(x).to(tu.get_device())
                y = filterbank(x)
                self.assertEqual(y.shape, (1, channels, int(sample_rate * 0.1), n_mel))
                self.assertEqual(type(y), np.ndarray if module == np else Tensor)

    def test_to_freq(self):
        for module, channels, sample_rate, n_freq, n_mel in self.params:
            with self.subTest(
                module=module,
                channels=channels,
                sample_rate=sample_rate,
                n_freq=n_freq,
                n_mel=n_mel,
            ):
                filterbank = tu.MelInverseFilterbank(sample_rate, n_freq, n_mel)
                x = np.ones((1, channels, int(sample_rate * 0.1), n_mel), dtype=complex)
                x = x if module == np else torch.from_numpy(x).to(tu.get_device())
                y = filterbank(x)
                self.assertEqual(y.shape, (1, channels, int(sample_rate * 0.1), n_freq))
                self.assertEqual(type(y), np.ndarray if module == np else Tensor)


class TestAudio(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.modules = (np, torch)

    def test_db(self):
        eps_list = (1e-12, 1e-3, 10)
        for eps in eps_list:
            with self.subTest(eps=eps):
                x = np.random.uniform(0, 1, 100)
                y = tu.db(x, eps)
                x_hat = tu.invert_db(y, eps)
                self.assertTrue(np.allclose(x, x_hat))

    def test_db_tensor(self):
        x = torch.rand(100)
        y = tu.db(x)
        x_hat = tu.invert_db(y)
        self.assertTrue(torch.allclose(x, x_hat))

    def test_power(self):
        x = np.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.power(x), 100)

    def test_power_multichannel(self):
        x = np.ones((2, 100))
        x[:, 50:] = -1
        y = tu.power(x)
        self.assertAlmostEqual(y[0], 100)
        self.assertAlmostEqual(y[1], 100)

    def test_power_tensor(self):
        x = torch.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.power(x), 100)

    def test_power_multichannel_tensor(self):
        x = torch.ones((2, 100))
        x[:, 50:] = -1
        y = tu.power(x)
        self.assertAlmostEqual(y[0], 100)
        self.assertAlmostEqual(y[1], 100)

    def test_energy(self):
        x = np.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.energy(x), 1)

    def test_energy_multichannel(self):
        x = np.ones((2, 100))
        x[:, 50:] = -1
        y = tu.energy(x)
        self.assertAlmostEqual(y[0], 1)
        self.assertAlmostEqual(y[1], 1)

    def test_energy_tensor(self):
        x = torch.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.energy(x), 1)

    def test_energy_multichannel_tensor(self):
        x = torch.ones((2, 100))
        x[:, 50:] = -1
        y = tu.energy(x)
        self.assertAlmostEqual(y[0], 1)
        self.assertAlmostEqual(y[1], 1)

    def test_rms(self):
        x = np.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.rms(x), 1)

    def test_rms_multichannel(self):
        x = np.ones((2, 100))
        x[:, 50:] = -1
        y = tu.rms(x)
        self.assertAlmostEqual(y[0], 1)
        self.assertAlmostEqual(y[1], 1)

    def test_rms_tensor(self):
        x = torch.ones(100)
        x[50:] = -1
        self.assertAlmostEqual(tu.rms(x), 1)

    def test_rms_multichannel_tensor(self):
        x = torch.ones((2, 100))
        x[:, 50:] = -1
        y = tu.rms(x)
        self.assertAlmostEqual(y[0], 1)
        self.assertAlmostEqual(y[1], 1)

    def test_snr(self):
        x = np.ones(10) * 10
        noise = np.ones(10) * 1
        self.assertAlmostEqual(tu.snr(x, noise), 20)

    def test_snr_tensor(self):
        x = torch.ones(10) * 10
        noise = torch.ones(10) * 1
        self.assertAlmostEqual(tu.snr(x, noise), 20)

    def test_fade_sides_1d(self):
        x = np.ones(200)
        x = tu.fade_sides(x)
        self.assertEqual(x[0], 0)
        self.assertEqual(x[-1], 0)

    def test_fade_sides_3d(self):
        x = np.ones((1, 2, 200))
        x = tu.fade_sides(x)
        self.assertEqual(x[0, 0, 0], 0)
        self.assertEqual(x[0, 0, -1], 0)

    def test_tensor_fade_sides_1d(self):
        x = torch.ones(200)
        x = tu.fade_sides(x)
        self.assertEqual(x[0], 0)
        self.assertEqual(x[-1], 0)

    def test_tensor_fade_sides_3d(self):
        x = np.ones((1, 2, 200))
        x = tu.fade_sides(x)
        self.assertEqual(x[0, 0, 0], 0)
        self.assertEqual(x[0, 0, -1], 0)

    def test_trim(self):
        x = np.zeros((1, 1, 5 * 160))
        y = tu.random_trim(x, 160, 2)
        self.assertEqual(y.shape, (1, 1, 2 * 160))

    def test_tensor_trim(self):
        x = torch.zeros((1, 1, 5 * 160))
        y = tu.random_trim(x, 160, 2)
        self.assertEqual(y.shape, (1, 1, 2 * 160))

    def test_trim_silence(self):
        margins = (0, 10)
        values = (1, 0.1, 100)
        params = product(self.modules, margins, values)
        for module, marg, val in params:
            x = module.zeros(100)
            x[30:40] = 0.1
            with self.subTest(module=module, marg=marg, val=val):
                y = tu.trim_silence(x, margin=marg)
                self.assertEqual(y.shape[-1], 10 + 2 * marg)

    def test_interleave(self):
        freqs = 10
        dims = (1, 4)
        params = product(self.modules, dims)
        for p in params:
            m, d = p
            with self.subTest(p=p):
                x_size = [freqs if i == (d - 1) else 1 for i in range(d)]

                # input
                if m == np:
                    x = np.random.normal(size=x_size)
                else:
                    x = torch.randn(x_size)

                xs = [x for _ in range(d)]
                y = tu.interleave(*xs)
                y_dim = x_size
                y_dim[-1] *= d
                self.assertEqual(y.shape, tuple(y_dim))
                for i, z in enumerate(xs):
                    e = m.abs(y[..., i::d] - z).max()
                    self.assertLess(e, 1e-6)

    def test_pack_audio_sequences(self):
        channels = (1, 2)
        length = (1.5, 2)
        sample_rate = (16000, 48000)
        tensor = (False, True)
        delete_last = (False, True)
        num_workers = (1, 4)
        params = product(
            channels, length, sample_rate, tensor, delete_last, num_workers
        )
        n_files = 6
        for p in params:
            (c, l, sr, t, d, w) = p
            filename = "mono" if c == 1 else "stereo"
            files = [get_test_data_dir() / f"{filename}.wav" for _ in range(n_files)]
            with self.subTest(p=p):
                itr = tu.pack_audio_sequences(files, l, sr, c, t, d, w)
                for i, seq in enumerate(itr):
                    self.assertEqual(seq.shape, (c, int(sr * l)))
                self.assertEqual(i + 1 if d else i, int(n_files / l))


if __name__ == "__main__":
    unittest.main(verbosity=2)



from typing import Iterator, List, Dict, Tuple
from torch import autograd as AG
import torch.nn.functional as F
from torch import nn, Tensor
from pathlib import Path
import numpy as np
import torch
import yaml

# export list
__all__ = [
    # pytorch utilities
    "enable_anomaly_detection",
    "load_model",
    "load_checkpoints",
    "sort_checkpoints",
    "get_submodules",
    "compute_gradients",
    "get_gradients",
    "get_model_parameters",
    "freeze_model",
    "quantize",
    "one_hot_quantization",
    "invert_one_hot",
    "CosineScheduler",
]


from torch_utilities.common import get_device


def enable_anomaly_detection():
    """
    Enable PyTorch anomaly detection
    """
    from torch import autograd

    autograd.anomaly_mode.set_detect_anomaly(True)


def load_model(
    model_path: Path,
    model_class: nn.Module,
    *model_args: List,
    **model_kwargs: Dict,
) -> nn.Module:
    """
    Creates a model from its last checkpoint.

    Parameters
    ----------
    model_path : Path
        Basepath to the model
    model_class : nn.Module
        Class of the model

    Returns
    -------
    nn.Module
        Loaded model
    """
    # loading from best checkpoint
    checkpoint_dir = model_path / "checkpoints"
    checkpoint_scores = checkpoint_dir / "ckpt.yml"
    checkpoint_scores = load_checkpoints(checkpoint_scores)
    best_checkpoint = checkpoint_dir / checkpoint_scores[0][0]
    config = model_path / "config.yml"
    model_state = torch.load(best_checkpoint, map_location=get_device())
    model_state = model_state["model_state"]

    # initializing the model
    model = model_class(config, *model_args, **model_kwargs)
    model.load_state_dict(model_state)
    model.eval()
    return model


def load_checkpoints(score_file: Path) -> List[Tuple[str, float]]:
    """
    Loads the ckpt.yml content ordered by the score.

    Parameters
    ----------
    score_file : Path
        Path to ckpt.yml

    Returns
    -------
    List[Tuple[str, float]]
        List of (checkpoint name, checkpoint score)
    """
    # loading
    with open(score_file) as f:
        checkpoints = yaml.unsafe_load(f)

    # ordering
    checkpoints = sort_checkpoints(checkpoints)
    return checkpoints


def sort_checkpoints(checkpoints: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
    """
    Sorts the checkpoint data structure based on the scores.

    Parameters
    ----------
    checkpoints : List[Tuple[str, float]]
        Checkpoint data structure, loaded with load_checkpoints.

    Returns
    -------
    List[Tuple[str, float]]
        Ordered checkpoints
    """
    _order = lambda t: -t[1]
    checkpoints = sorted(checkpoints, key=_order)
    return checkpoints


def get_submodules(model: nn.Module) -> List[nn.Module]:
    """
    Gets all the submodules without children from a model.

    Parameters
    ----------
    model : nn.Module
        Target model

    Returns
    -------
    List[nn.Module]
        Model submodules
    """
    modules = [x for x in model.modules() if len(list(x.children())) == 0]
    return modules


def get_gradients(model: nn.Module) -> Tensor:
    """
    Gets a model gradients norm through its submodules.

    Parameters
    ----------
    model : nn.Module
        Target model

    Returns
    -------
    Tensor
        Model gradients
    """
    modules = get_submodules(model)

    valid = (
        lambda x, n: hasattr(x, n)
        and isinstance(getattr(x, n), Tensor)
        and getattr(x, n).grad is not None
    )

    _norm = lambda x: torch.linalg.norm(x).item()
    f = lambda n: [_norm(getattr(x, n).grad) if valid(x, n) else 0 for x in modules]
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    w_grad = f("weight")
    b_grad = f("bias")
    # weight normalization
    g_grad = f("weight_g")
    v_grad = f("weight_v")
    # GRU ~ ~ ~ ~ ~ ~ ~ ~
    max_gru = 4
    gru_w_ih = [f(f"weight_ih_l{i}") for i in range(max_gru)]
    gru_w_hh = [f(f"weight_hh_l{i}") for i in range(max_gru)]
    bias_w_ih = [f(f"bias_ih_l{i}") for i in range(max_gru)]
    bias_w_hh = [f(f"bias_hh_l{i}") for i in range(max_gru)]
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
    zipped = zip(
        w_grad, b_grad, g_grad, v_grad, *gru_w_ih, *gru_w_hh, *bias_w_ih, *bias_w_hh
    )
    grad = [sum(xs) for xs in zipped]
    grad = torch.FloatTensor(grad)
    return grad


def compute_gradients(x: Tensor, y: Tensor, keep_graph: bool = True) -> Tensor:
    """
    Computes the matrix-vector product between the Jacobian calculated from the
    function with y as output and x as input and a vector of ones of the same
    shape as y.

    Parameters
    ----------
    x : Tensor
        Input of the target function, must require grad
    y : Tensor
        Output of the target function
    keep_graph : bool, optional
        If True the result can be used in further autograd graphs, it can then be used
        for higher degree derivative optimization (i.e. as an derivative optimization term in a loss
        such in https://arxiv.org/abs/1704.00028), by default True

    Returns
    -------
    Tensor
        matrix-vector product between the Jacobian
    """
    ones = torch.ones_like(y)
    grad = AG.grad(y, x, ones, retain_graph=keep_graph, create_graph=keep_graph)[0]
    return grad


def get_model_parameters(model: nn.Module) -> int:
    """
    Gets the total number of parameters (trainable and untrainable).

    Parameters
    ----------
    model : nn.Module
        Target model

    Returns
    -------
    int
        Total number of parameters
    """
    return sum(p.numel() for p in model.parameters())


def freeze_model(model: nn.Module) -> None:
    """
    Freezed the weights of a target module.

    Parameters
    ----------
    model : nn.Module
        Model to freeze
    """
    for p in model.parameters():
        p.requires_grad = False


def quantize(x: Tensor, steps: int, min: float = -1, max: float = 1) -> Tensor:
    """
    Quantize a real signal.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (B, ...)
    steps : int
        Number of quantization steps
    min : float, optional
        Minimum value of the input (mapped to 0), by default -1
    max : float, optional
        Maximum value of the input (mapped to steps-1), by default 1

    Returns
    -------
    Tensor
        Quantized tensor of shape (B, ...)
    """
    x = (x - min) / (max - min) * steps
    x = x.floor().clip(0, steps - 1).to(int)
    return x


def one_hot_quantization(
    x: Tensor, steps: int, min: float = -1, max: float = 1
) -> Tensor:
    """
    Quantize a real signal and applies a one-hot vector transform.

    Parameters
    ----------
    x : Tensor
        Input tensor of shape (B, ...)
    steps : int
        Number of quantization steps
    min : float, optional
        Minimum value of the input (mapped to 0), by default -1
    max : float, optional
        Maximum value of the input (mapped to steps-1), by default 1

    Returns
    -------
    Tensor
        One-hot-quantized tensor of shape (B, steps, ...)
    """
    n = len(x.shape)
    dims = list(range(n))
    dims.insert(1, n)

    # quantization
    x = quantize(x, steps, min, max)

    # one-hot vector
    x = F.one_hot(x, steps)
    x = x.permute(dims)

    return x


def invert_one_hot(x: Tensor) -> Tensor:
    """
    Transforms a one-hot tensor to a integer label representation.

    Parameters
    ----------
    x : Tensor
        On-hot tensor of shape (B, steps, ...)

    Returns
    -------
    Tensor
        Label representation of shape (B, ...)
    """
    x = torch.argmax(x, dim=1)
    return x


class CosineScheduler:
    def __init__(
        self,
        start_value: float,
        final_value: float,
        total_epochs: int,
        iterations_per_epoch: int = 1,
        warmup_epochs: int = 0,
    ) -> None:
        """
        Implements a single cycle cosine scheduler.

        Parameters
        ----------
        start_value : float
            Initial value
        final_value : float
            Last value
        total_epochs : int
            Total number of epochs
        iterations_per_epoch : int, optional
            Iterations for each eopoch, by default 1
        warmup_epochs : int, optional
            Number of warmup epochs (linear rise to the start_value), by default 0
        """
        self.start_value = start_value
        self.final_value = final_value
        self.total_epochs = total_epochs
        self.iterations_per_epoch = iterations_per_epoch
        self.warmup_epochs = warmup_epochs

        self.schedule = self._compute_schedule()
        self.current_index = 0

    def __iter__(self) -> Iterator[float]:
        return iter(self.schedule.tolist())

    def __next__(self) -> float:
        value = self.schedule[self.current_index]
        self.current_index += 1
        return value

    def reset(self) -> None:
        """
        Resets the scheduler state.
        """
        self.current_index = 0

    def _compute_schedule(self) -> np.ndarray:
        """
        Computes the full schedule.

        Returns
        -------
        np.ndarray
            Schedule array
        """
        warmup_schedule = np.array([])
        warmup_iters = self.warmup_epochs * self.iterations_per_epoch
        if self.warmup_epochs > 0:
            warmup_start = self.start_value / warmup_iters
            warmup_schedule = np.linspace(warmup_start, self.start_value, warmup_iters)

        iters_after_warmup = (
            self.total_epochs * self.iterations_per_epoch - warmup_iters
        )
        cycle_lengths = [iters_after_warmup]

        schedule_cycles = []

        # supporting one single cycle
        num_cycles = 1
        for i in range(num_cycles):
            iters = np.arange(cycle_lengths[i])
            schedule = np.array(
                [
                    self.final_value
                    + 0.5
                    * (self.start_value - self.final_value)
                    * (1 + np.cos(np.pi * i / (len(iters))))
                    for i in iters
                ]
            )
            schedule_cycles.append(schedule)

        schedule = np.concatenate((warmup_schedule, *schedule_cycles))
        schedule = schedule[: self.total_epochs * self.iterations_per_epoch]

        assert len(schedule) == self.total_epochs * self.iterations_per_epoch
        return schedule



from torch.nn.utils import weight_norm
from typing import Tuple, Type
from itertools import product
from torch import Tensor
from torch import nn
import numpy as np
import unittest
import torch

import torch_utilities as tu
from torch_utilities import repeat_test, set_device
from torch_utilities.modules import (
    get_causal_longformer_mask,
    get_time_value,
    get_freq_value,
    LambdaLayer,
    Lookahead,
    Reparameterize,
    ScaleChannels2d,
    UnfoldSpectrogram,
    FoldSpectrogram,
    ResidualWrap,
    GroupedLinear,
    CausalConv2d,
    CausalSubConv2d,
    CausalConv2dNormAct,
    CausalSmoothedTConv,
    DenseConvBlock,
    GruNormAct,
    SlidingCausalMultiheadAttention,
)

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(876)
    set_device("auto")
    torch.set_grad_enabled(False)


def _get_input(in_channels: int, in_freqs: int, dtype: Type) -> Tuple:
    batch_size = 1
    frames = 30
    x = torch.randn(batch_size, in_channels, frames, in_freqs, dtype=dtype)
    return x


_get_f = lambda x: x if isinstance(x, int) else x[1]

sum_merge = LambdaLayer(lambda x, y: x + y)

# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestLookahead(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.x = torch.zeros(1, 2, 10, 16)

    def test_no_maintain_shape(self):
        lookahead = Lookahead(4)
        y = lookahead(self.x)
        self.assertEqual(y.shape, (1, 2, 6, 16))

    def test_maintain_shape(self):
        lookahead = Lookahead(4, maintain_shape=True)
        y = lookahead(self.x)
        self.assertEqual(y.shape, self.x.shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestUnfoldFoldSpectrogram(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_channels = 2
        self.in_frames = 20
        self.in_num_blocks = 3
        self.in_freqs = 4

        self.block_size = (2, 4)
        self.stride = (1, 2)
        self.params = product(self.block_size, self.stride)

    def get_instance(self, p: Tuple, unfold: bool) -> CausalConv2d:
        (block_size, stride) = p
        if unfold:
            instance = UnfoldSpectrogram(
                block_size=block_size,
                stride=stride,
            )
        else:
            instance = FoldSpectrogram(
                block_size=block_size,
                stride=stride,
                channels=self.in_channels,
            )
        return instance

    def get_input(self, p: Tuple, unfold: bool) -> Tensor:
        (block_size, stride) = p

        if unfold:
            x = torch.randn((1, self.in_channels, self.in_frames, self.in_freqs))
        else:
            x = torch.randn(
                (1, self.in_channels * self.in_num_blocks, block_size, self.in_freqs)
            )
        return x

    def test_unfold(self):
        for p in self.params:
            (block_size, stride) = p
            with self.subTest(p=p):
                unfold = self.get_instance(p, unfold=True)
                x = self.get_input(p, unfold=True)
                y = unfold(x)

                n_blocks = (self.in_frames - (block_size - 1) - 1) / stride + 1
                expected_shape = (
                    1,
                    self.in_channels * n_blocks,
                    block_size,
                    self.in_freqs,
                )
                self.assertTrue(y.shape, expected_shape)

    def test_unfold_assertion(self):
        x = torch.ones(1, 1, 11, 4)
        unfold = UnfoldSpectrogram(4, 2)
        with self.assertRaises(AssertionError):
            unfold(x)

    def test_fold(self):
        for p in self.params:
            (block_size, stride) = p
            with self.subTest(p=p):
                fold = self.get_instance(p, unfold=False)
                x = self.get_input(p, unfold=False)
                y = fold(x)
                self.assertEqual(y.shape[1], self.in_channels)
                self.assertEqual(y.shape[3], self.in_freqs)

    def test_inversion(self):
        for p in self.params:
            (block_size, stride) = p
            with self.subTest(p=p):
                unfold = self.get_instance(p, unfold=True)
                fold = self.get_instance(p, unfold=False)
                x = self.get_input(p, unfold=True)
                y = unfold(x)
                x_hat = fold(y)
                e = (x - x_hat).abs().max().item()
                self.assertLess(e, 1e-6)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestResidualWrap(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.x = torch.ones(1, 2, 10, 16)

    def test_single_layer(self):
        layer = ResidualWrap(nn.Identity())
        y = layer(self.x)
        self.assertEqual(y.shape, self.x.shape)
        self.assertLess((self.x * 2 - y).abs().max(), 1e-6)

    def test_multiple_layers(self):
        layer = ResidualWrap(nn.Identity(), nn.Identity())
        y = layer(self.x)
        self.assertEqual(y.shape, self.x.shape)
        self.assertLess((self.x * 2 - y).abs().max(), 1e-6)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestCausalConv2d(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_channels = (1, 2)
        self.out_channels = (1, 3)
        self.kernel_size = (1, 4, (2, 3))
        self.stride_f = (1, 2, 4)
        self.padding_f = (None, 0, 1, 2)
        self.dilation = (1, 2, 4, (3, 2))
        self.bias = (False, True)
        self.separable = (False, True)
        self.enable_weight_norm = (False, True)
        self.dtype = (torch.float, torch.double)
        self.in_freqs = (32, 33)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.stride_f,
            self.padding_f,
            self.dilation,
            self.bias,
            self.separable,
            self.enable_weight_norm,
            self.dtype,
            self.in_freqs,
        )

    def get_instance(self, p: Tuple) -> CausalConv2d:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride_f,
            padding_f,
            dilation,
            bias,
            separable,
            enable_weight_norm,
            dtype,
            in_freqs,
        ) = p
        instance = CausalConv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride_f=stride_f,
            padding_f=padding_f,
            dilation=dilation,
            bias=bias,
            separable=separable,
            enable_weight_norm=enable_weight_norm,
            dtype=dtype,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride_f,
            padding_f,
            dilation,
            bias,
            separable,
            enable_weight_norm,
            dtype,
            in_freqs,
        ) = p
        x = _get_input(in_channels, in_freqs, dtype)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride_f,
                padding_f,
                dilation,
                bias,
                separable,
                enable_weight_norm,
                dtype,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                conv = self.get_instance(p)
                self.assertEqual(type(conv.layers), nn.Sequential)
                padding_layer = (
                    nn.Identity if padding_f is not None else nn.ConstantPad2d
                )
                if separable:
                    self.assertEqual(type(conv.layers[0]), nn.ConstantPad2d)
                    self.assertEqual(type(conv.layers[1]), padding_layer)
                    self.assertEqual(type(conv.layers[2]), nn.Conv2d)
                    self.assertEqual(type(conv.layers[3]), nn.Conv2d)
                    self.assertTrue(
                        conv.layers[2].kernel_size == kernel_size
                        or conv.layers[2].kernel_size == (kernel_size, kernel_size)
                    )
                    self.assertEqual(conv.layers[3].kernel_size, (1, 1))
                else:
                    self.assertEqual(type(conv.layers[0]), nn.ConstantPad2d)
                    self.assertEqual(type(conv.layers[1]), padding_layer)
                    self.assertEqual(type(conv.layers[2]), nn.Conv2d)
                    self.assertTrue(
                        conv.layers[2].kernel_size == kernel_size
                        or conv.layers[2].kernel_size == (kernel_size, kernel_size)
                    )

                if enable_weight_norm:
                    self.assertEqual(conv._normalize, weight_norm)

    def test_forward(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride_f,
                padding_f,
                dilation,
                bias,
                separable,
                enable_weight_norm,
                dtype,
                in_freqs,
            ) = p

            # skipping autopadding when stride_f doesn't divide in_freqs
            if padding_f is None and in_freqs % stride_f != 0:
                continue

            with self.subTest(p=p):
                conv = self.get_instance(p)
                x = self.get_input(p)
                y = conv(x)
                batch_size, _, frames = x.shape[:3]
                dilation_f = _get_f(dilation)
                kernel_f = _get_f(kernel_size)

                if padding_f is None:
                    # auto padding
                    out_freqs = in_freqs // stride_f
                else:
                    # manual padding
                    out_freqs = (
                        in_freqs + 2 * padding_f - dilation_f * (kernel_f - 1) - 1
                    )
                    out_freqs = int(out_freqs / stride_f + 1)

                expected_shape = (batch_size, out_channels, frames, out_freqs)
                self.assertEqual(y.shape, expected_shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestCausalConv2dNormAct(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_channels = (1, 2)
        self.out_channels = (1, 3)
        self.kernel_size = (1, 4, (2, 3))
        self.stride_f = (2, 4)
        self.dilation = (1, 4, (3, 2))
        self.separable = (False, True)
        self.activation = (None, nn.ReLU())
        self.residual_merge = (None, sum_merge)
        self.disable_batchnorm = (False, True)
        self.enable_weight_norm = (False, True)
        self.merge_after_conv = (False, True)
        self.dtype = (torch.float, torch.double)
        self.in_freqs = (32, 64)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.stride_f,
            self.dilation,
            self.separable,
            self.activation,
            self.residual_merge,
            self.disable_batchnorm,
            self.enable_weight_norm,
            self.merge_after_conv,
            self.dtype,
            self.in_freqs,
        )

    def get_instance(self, p: Tuple) -> CausalConv2dNormAct:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride_f,
            dilation,
            separable,
            activation,
            residual_merge,
            disable_batchnorm,
            enable_weight_norm,
            merge_after_conv,
            dtype,
            in_freqs,
        ) = p

        if in_freqs % 2 == 1:
            stride_f = 1

        instance = CausalConv2dNormAct(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride_f=stride_f,
            dilation=dilation,
            separable=separable,
            activation=activation,
            residual_merge=residual_merge,
            disable_batchnorm=disable_batchnorm,
            enable_weight_norm=enable_weight_norm,
            merge_after_conv=merge_after_conv,
            dtype=dtype,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride_f,
            dilation,
            separable,
            activation,
            residual_merge,
            disable_batchnorm,
            enable_weight_norm,
            merge_after_conv,
            dtype,
            in_freqs,
        ) = p
        x = _get_input(in_channels, in_freqs, dtype)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride_f,
                dilation,
                separable,
                activation,
                residual_merge,
                disable_batchnorm,
                enable_weight_norm,
                merge_after_conv,
                dtype,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                conv = self.get_instance(p)
                causal_conv_2d = conv.conv
                self.assertEqual(type(causal_conv_2d), CausalConv2d)
                if enable_weight_norm:
                    disable_batchnorm = True
                    self.assertEqual(causal_conv_2d._normalize, weight_norm)
                if activation is None:
                    self.assertEqual(type(conv.activation), nn.Identity)
                batchnorm = conv.batchnorm
                expected_batchnorm = (
                    nn.Identity if disable_batchnorm else nn.BatchNorm2d
                )
                self.assertEqual(type(batchnorm), expected_batchnorm)

    def test_forward(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride_f,
                dilation,
                separable,
                activation,
                residual_merge,
                disable_batchnorm,
                enable_weight_norm,
                merge_after_conv,
                dtype,
                in_freqs,
            ) = p
            if in_freqs % 2 == 1:
                stride_f = 1

            if (not merge_after_conv) and (stride_f != 1):
                # excluding cases where the merge is before the stride
                # and the stride is different than 1
                return

            with self.subTest(p=p):
                conv = self.get_instance(p)
                x = self.get_input(p)
                y = conv(x)
                batch_size, _, frames = x.shape[:3]
                out_freqs = in_freqs // stride_f
                expected_shape = (batch_size, out_channels, frames, out_freqs)
                self.assertEqual(y.shape, expected_shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestReparameterize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.reparam = Reparameterize()
        self.eps = 1e-1

    @repeat_test(5)
    def test_zero_mean(self):
        mu = torch.zeros(100000)
        logvar = torch.ones(100000)
        y = self.reparam(mu, logvar)
        y = torch.mean(y)
        self.assertLess(y.item(), self.eps)

    @repeat_test(5)
    def test_nonzero_mean(self):
        mu = torch.ones(100000)
        logvar = torch.ones(100000)
        y = self.reparam(mu, logvar)
        y = torch.mean(y) - 1
        self.assertLess(y.item(), self.eps)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestScaleChannels2d(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.scale = ScaleChannels2d(2)

    def set_weights(self, weights: list):
        w = self.scale.scale.weight.data
        self.scale.scale.weight.data = torch.ones_like(w)
        for i, w in enumerate(weights):
            self.scale.scale.weight.data[i] = w

    def test_scaling(self):
        self.set_weights([1, 2])
        x = torch.ones((1, 2, 10, 32))
        y = self.scale(x)
        self.assertTrue(torch.allclose(y[0, 0], torch.ones_like(y[0, 0])))
        self.assertTrue(torch.allclose(y[0, 1], torch.ones_like(y[0, 1]) * 2))


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestCausalSmoothedTConv(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_channels = (1,)
        self.out_channels = (3,)
        self.post_conv_kernel_size = (1, 2, (1, 3))
        self.post_conv_count = (1, 2)
        self.post_conv_dilation = (None, 1, 3)
        self.tconv_stride_f = (1, 2)
        self.separable = (False, True)
        self.disable_batchnorm = (False, True)
        self.enable_weight_norm = (False, True)
        self.activation = (None, nn.ReLU())
        self.residual_merge = (None, sum_merge)
        self.dtype = (torch.float,)
        self.in_freqs = (50,)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.in_channels,
            self.out_channels,
            self.post_conv_kernel_size,
            self.post_conv_count,
            self.post_conv_dilation,
            self.tconv_stride_f,
            self.separable,
            self.disable_batchnorm,
            self.enable_weight_norm,
            self.activation,
            self.residual_merge,
            self.dtype,
            self.in_freqs,
        )

    def get_instance(self, p: Tuple) -> CausalSmoothedTConv:
        (
            in_channels,
            out_channels,
            post_conv_kernel_size,
            post_conv_count,
            post_conv_dilation,
            tconv_stride_f,
            separable,
            disable_batchnorm,
            enable_weight_norm,
            activation,
            residual_merge,
            dtype,
            in_freqs,
        ) = p

        instance = CausalSmoothedTConv(
            in_channels=in_channels,
            out_channels=out_channels,
            post_conv_kernel_size=post_conv_kernel_size,
            post_conv_count=post_conv_count,
            post_conv_dilation=post_conv_dilation,
            tconv_stride_f=tconv_stride_f,
            separable=separable,
            disable_batchnorm=disable_batchnorm,
            enable_weight_norm=enable_weight_norm,
            activation=activation,
            residual_merge=residual_merge,
            dtype=dtype,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            in_channels,
            out_channels,
            post_conv_kernel_size,
            post_conv_count,
            post_conv_dilation,
            tconv_stride_f,
            separable,
            disable_batchnorm,
            enable_weight_norm,
            activation,
            residual_merge,
            dtype,
            in_freqs,
        ) = p
        x = _get_input(in_channels, in_freqs, dtype)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                post_conv_kernel_size,
                post_conv_count,
                post_conv_dilation,
                tconv_stride_f,
                separable,
                disable_batchnorm,
                enable_weight_norm,
                activation,
                residual_merge,
                dtype,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                upsamp = self.get_instance(p)

                # tconv
                tconv = upsamp.tconv
                self.assertEqual(tconv.in_channels, in_channels)
                self.assertEqual(tconv.out_channels, out_channels)
                self.assertEqual(tconv.kernel_size, (1, 2 * tconv_stride_f))

                # post conv
                pconv = upsamp.conv
                self.assertEqual(type(pconv), nn.Sequential)
                self.assertEqual(len(pconv), post_conv_count)
                for i, c in enumerate(pconv):
                    if isinstance(c, CausalConv2d) and not separable:
                        self.assertEqual(c.in_channels, out_channels)
                        self.assertEqual(c.out_channels, out_channels)
                        self.assertEqual(c.separable, separable)
                        self.assertEqual(c.enable_weight_norm, enable_weight_norm)
                        self.assertTrue(
                            c.kernel_size == post_conv_kernel_size
                            or c.kernel_size
                            == (post_conv_kernel_size, post_conv_kernel_size)
                        )
                        if post_conv_dilation is None:
                            k_t = get_time_value(c.kernel_size)
                            k_f = get_freq_value(c.kernel_size)
                            expected_dilation = (k_t**i, k_f**i)
                            self.assertEqual(c.dilation, expected_dilation)

                # batchnorm
                if enable_weight_norm:
                    self.assertTrue(upsamp.disable_batchnorm)

                # activation
                act = upsamp.activation
                if activation is None:
                    self.assertEqual(type(act), nn.Identity)
                else:
                    self.assertEqual(type(act), type(activation))

    def test_forward(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                post_conv_kernel_size,
                post_conv_count,
                post_conv_dilation,
                tconv_stride_f,
                separable,
                disable_batchnorm,
                enable_weight_norm,
                activation,
                residual_merge,
                dtype,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                upsamp = self.get_instance(p)
                x = self.get_input(p)
                y = upsamp(x)
                batch_size, _, frames = x.shape[:3]
                out_freqs = tconv_stride_f * in_freqs
                expected_shape = (batch_size, out_channels, frames, out_freqs)
                self.assertEqual(y.shape, expected_shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestGroupedLinear(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass

    def test_assert(self):
        params = (
            (65, 32),
            (64, 31),
        )
        for p in params:
            with self.subTest(p=p):
                with self.assertRaises(AssertionError):
                    GroupedLinear(p[0], p[1], groups=8)

    def test_no_groups(self):
        x = torch.rand((1, 10, 32))
        gl = GroupedLinear(32, 64, 1)
        lin = nn.Linear(32, 64, bias=False)
        gl.weight.data = torch.ones_like(gl.weight.data)
        lin.weight.data = torch.ones_like(lin.weight.data)
        self.assertTrue(torch.allclose(gl(x), lin(x)))


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestGruNormAct(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_size = (16, 32)
        self.h_size = (8, 64)
        self.merge = (None, (lambda x, y: y))
        self.activation = (None, nn.Sigmoid())
        self.batchnorm = (True, False)
        self.params = product(
            self.in_size,
            self.h_size,
            self.merge,
            self.activation,
            self.batchnorm,
        )

    def get_input(self, in_size: int) -> Tensor:
        x = torch.randn((1, 10, in_size))
        return x

    def test_submodules(self):
        for p in self.params:
            in_size, h_size, merge, act, batchnorm = p
            with self.subTest(p=p):
                gru = GruNormAct(
                    input_size=in_size,
                    hidden_size=h_size,
                    residual_merge=merge,
                    activation=act,
                    disable_batchnorm=not batchnorm,
                )
                submodules = set(type(m) for m in tu.get_submodules(gru))
                batchnorm_type = nn.BatchNorm1d if batchnorm else nn.Identity
                act_type = nn.Identity if act is None else type(act)
                expected = set((nn.GRU, batchnorm_type, act_type))
                self.assertEqual(expected, submodules)

    def test_forward(self):
        for p in self.params:
            in_size, h_size, merge, act, batchnorm = p
            with self.subTest(p=p):
                x = self.get_input(in_size)
                gru = GruNormAct(
                    input_size=in_size,
                    hidden_size=h_size,
                    residual_merge=merge,
                    activation=act,
                    disable_batchnorm=not batchnorm,
                )
                y, h = gru(x)
                self.assertEqual(y.shape, (*x.shape[:-1], h_size))
                self.assertEqual(h.shape, (x.shape[0], 1, h_size))


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
class TestDenseConvBlock(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.channels = (2, 4)
        self.kernel_size = (3, (3, 5))
        self.dilation = (None, 1, (2, 5))
        self.disable_dilation_f = (False, True)
        self.depth = (1, 4)
        self.final_stride = (1, 2)
        self.disable_layernorm = (False, True)
        self.enable_weight_norm = (False, True)
        self.activation = (None, nn.ReLU())
        self.feature_size = (40,)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.channels,
            self.kernel_size,
            self.dilation,
            self.disable_dilation_f,
            self.depth,
            self.final_stride,
            self.disable_layernorm,
            self.enable_weight_norm,
            self.activation,
            self.feature_size,
        )

    def get_instance(self, p: Tuple) -> DenseConvBlock:
        (
            channels,
            kernel_size,
            dilation,
            disable_dilation_f,
            depth,
            final_stride,
            disable_layernorm,
            enable_weight_norm,
            activation,
            feature_size,
        ) = p
        disable_layernorm = disable_layernorm or enable_weight_norm
        instance = DenseConvBlock(
            channels=channels,
            kernel_size=kernel_size,
            feature_size=feature_size,
            dilation=dilation,
            disable_dilation_f=disable_dilation_f,
            depth=depth,
            final_stride=final_stride,
            disable_layernorm=disable_layernorm,
            enable_weight_norm=enable_weight_norm,
            activation=activation,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            channels,
            kernel_size,
            dilation,
            disable_dilation_f,
            depth,
            final_stride,
            disable_layernorm,
            enable_weight_norm,
            activation,
            feature_size,
        ) = p
        x = _get_input(channels, feature_size, None)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                channels,
                kernel_size,
                dilation,
                disable_dilation_f,
                depth,
                final_stride,
                disable_layernorm,
                enable_weight_norm,
                activation,
                feature_size,
            ) = p
            disable_layernorm = disable_layernorm or enable_weight_norm
            with self.subTest(p=p):
                dcb = self.get_instance(p)
                layers = dcb.layers
                for i, seq in enumerate(layers):
                    expected_layernorm = (
                        nn.Identity if disable_layernorm else nn.LayerNorm
                    )
                    expected_layernorm = (
                        nn.Identity if i == (depth - 1) else expected_layernorm
                    )
                    expected_activation = (
                        nn.Identity if activation is None else type(activation)
                    )
                    self.assertEqual(type(seq), nn.Sequential)
                    self.assertEqual(type(seq[0]), CausalConv2dNormAct)
                    self.assertEqual(type(seq[1]), expected_layernorm)
                    self.assertEqual(type(seq[2]), expected_activation)
                    self.assertEqual(seq[0].in_channels, (i + 1) * channels)
                    self.assertEqual(seq[0].out_channels, channels)
                    self.assertEqual(
                        seq[0].stride_f, 1 if i != (depth - 1) else final_stride
                    )
                    self.assertTrue(seq[0].disable_batchnorm)
                    self.assertEqual(seq[0].enable_weight_norm, enable_weight_norm)

    def test_forward(self):
        for p in self.params:
            (
                channels,
                kernel_size,
                dilation,
                disable_dilation_f,
                depth,
                final_stride,
                disable_layernorm,
                enable_weight_norm,
                activation,
                feature_size,
            ) = p
            disable_layernorm = disable_layernorm or enable_weight_norm
            with self.subTest(p=p):
                dcb = self.get_instance(p)
                x = self.get_input(p)
                y = dcb(x)
                B, C, T, F = x.shape
                self.assertEqual(y.shape, (B, C, T, F // final_stride))


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestCausalSubConv2d(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.in_channels = (2,)
        self.out_channels = (3,)
        self.kernel_size = (4, (5, 3))
        self.stride = (1, 2, 4)
        self.dilation = (1, (2, 3))
        self.bias = (False,)
        self.separable = (False, True)
        self.enable_weight_norm = (False, True)
        self.upsampling_dim = ("time",)
        # self.upsampling_dim = ("freq", "time")
        self.in_freqs = (16,)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.in_channels,
            self.out_channels,
            self.kernel_size,
            self.stride,
            self.dilation,
            self.bias,
            self.separable,
            self.enable_weight_norm,
            self.upsampling_dim,
            self.in_freqs,
        )

    def get_instance(self, p: Tuple) -> CausalSubConv2d:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride,
            dilation,
            bias,
            separable,
            enable_weight_norm,
            upsampling_dim,
            in_freqs,
        ) = p
        instance = CausalSubConv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            bias=bias,
            separable=separable,
            upsampling_dim=upsampling_dim,
            enable_weight_norm=enable_weight_norm,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            in_channels,
            out_channels,
            kernel_size,
            stride,
            dilation,
            bias,
            separable,
            enable_weight_norm,
            upsampling_dim,
            in_freqs,
        ) = p
        x = _get_input(in_channels, in_freqs, None)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride,
                dilation,
                bias,
                separable,
                enable_weight_norm,
                upsampling_dim,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                csc = self.get_instance(p)
                self.assertEqual(len(csc.layers), stride)
                for layer in csc.layers:
                    self.assertEqual(layer.in_channels, in_channels)
                    self.assertEqual(layer.out_channels, out_channels)
                    self.assertEqual(layer.kernel_size, kernel_size)
                    self.assertEqual(layer.stride_f, 1)
                    self.assertEqual(layer.dilation, dilation)
                    self.assertEqual(layer.bias, bias)
                    self.assertEqual(layer.separable, separable)
                    self.assertEqual(layer.enable_weight_norm, enable_weight_norm)

    def test_forward(self):
        for p in self.params:
            (
                in_channels,
                out_channels,
                kernel_size,
                stride,
                dilation,
                bias,
                separable,
                enable_weight_norm,
                upsampling_dim,
                in_freqs,
            ) = p
            with self.subTest(p=p):
                csc = self.get_instance(p)
                x = self.get_input(p)
                y = csc(x)
                B, C, T, F = x.shape
                expected_shape = (
                    (B, out_channels, T, in_freqs * stride)
                    if upsampling_dim == "freq"
                    else (B, out_channels, T * stride, in_freqs)
                )
                self.assertEqual(y.shape, expected_shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =


class TestSlidingCausalMultiheadAttention(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.channels = (2,)
        self.sequence_len = (10,)
        self.embed_dim = (32, 64)
        self.stride = (5, 2)
        self.num_heads = (1,)
        self.dropout = (0,)
        self.bias = (True,)
        self.receptive_field = (None,)
        self.attn_mask = (None,)
        # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~
        self.params = product(
            self.channels,
            self.sequence_len,
            self.embed_dim,
            self.stride,
            self.num_heads,
            self.dropout,
            self.bias,
            self.receptive_field,
            self.attn_mask,
        )

    def get_instance(self, p: Tuple) -> SlidingCausalMultiheadAttention:
        (
            channels,
            sequence_len,
            embed_dim,
            stride,
            num_heads,
            dropout,
            bias,
            receptive_field,
            attn_mask,
        ) = p
        instance = SlidingCausalMultiheadAttention(
            channels=channels,
            sequence_len=sequence_len,
            embed_dim=embed_dim,
            stride=stride,
            num_heads=num_heads,
            dropout=dropout,
            bias=bias,
            receptive_field=receptive_field,
            attn_mask=attn_mask,
        )
        return instance

    def get_input(self, p: Tuple) -> Tensor:
        (
            channels,
            sequence_len,
            embed_dim,
            stride,
            num_heads,
            dropout,
            bias,
            receptive_field,
            attn_mask,
        ) = p
        x = _get_input(channels, embed_dim, None)
        return x

    def test_inner_modules(self):
        for p in self.params:
            (
                channels,
                sequence_len,
                embed_dim,
                stride,
                num_heads,
                dropout,
                bias,
                receptive_field,
                attn_mask,
            ) = p
            with self.subTest(p=p):
                att = self.get_instance(p)
                self.assertEqual(type(att.mh_attention), nn.MultiheadAttention)
                self.assertEqual(type(att.unfold), UnfoldSpectrogram)
                self.assertEqual(type(att.fold), FoldSpectrogram)

                # default attention mask
                if attn_mask is None and receptive_field is None:
                    expected = get_causal_longformer_mask(
                        sequence_len, sequence_len // 2
                    )
                    max_err = (expected - att.attn_mask).abs().max()
                    self.assertLess(max_err, 1e-12)

    def test_forward(self):
        for p in self.params:
            (
                channels,
                sequence_len,
                embed_dim,
                stride,
                num_heads,
                dropout,
                bias,
                receptive_field,
                attn_mask,
            ) = p
            with self.subTest(p=p):
                att = self.get_instance(p)
                x = self.get_input(p)
                y = att(x)
                self.assertEqual(x.shape, y.shape)


# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

if __name__ == "__main__":
    unittest.main(verbosity=2)



import torch.nn.functional as F
from torch import Tensor


from torch_utilities.common import split_complex, pack_complex
from torch_utilities.audio import stft, istft

# export list
__all__ = [
    "STFT",
    "ISTFT",
]


class STFT:
    def __init__(
        self,
        sample_rate: int,
        hopsize_ms: float,
        overlap_ratio: int = 2,
        win_oversamp: int = 1,
        pack_niquist: bool = True,
    ) -> None:
        """
        Apply a STFT to the input batch.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        hopsize_ms : int
            STFT hopsize in ms
        overlap_ratio : int, optional
            Overlap ratio between two frames
        win_oversamp : int, optional
            Zero padding oversampling applied to the window,
            by default 1 equals to no oversampling
        pack_niquist : bool, optional
            If True the niquist frequency is encoded as the imaginary part of
            the DC bin, by default True
        """
        self.sample_rate = sample_rate
        self.hopsize_ms = hopsize_ms
        self.overlap_ratio = overlap_ratio
        self.win_oversamp = win_oversamp
        self.pack_niquist = pack_niquist

    def __call__(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Signal of shape (B, C, T)

        Returns
        -------
        Tensor
            STFT of shape (B, 2*C, T', F)
        """
        batches, channels = x.shape[:2]
        x = x.flatten(0, 1)
        x = stft(
            x,
            sample_rate=self.sample_rate,
            hopsize_ms=self.hopsize_ms,
            win_len_ms=self.hopsize_ms * self.overlap_ratio,
            win_oversamp=self.win_oversamp,
        )
        x = x.unflatten(0, (batches, channels))
        if self.pack_niquist:
            # putting the niquist as imag of 0-th bin
            x[..., 0] += 1j * x[..., -1].real
            x = x[..., :-1]
        x = split_complex(x)
        return x


class ISTFT:
    def __init__(
        self,
        sample_rate: int,
        hopsize_ms: float,
        overlap_ratio: int = 2,
        win_oversamp: int = 1,
        unpack_niquist: bool = True,
    ) -> None:
        """
        Apply a ISTFT to the input batch.

        Parameters
        ----------
        sample_rate : int
            Sample rate of the signal
        hopsize_ms : int
            STFT hopsize in ms
        overlap_ratio : int, optional
            Overlap ratio between two frames
        win_oversamp : int, optional
            Zero padding oversampling applied to the window,
            by default 1 equals to no oversampling
        unpack_niquist : bool, optional
            If True the niquist frequency is decoded as the imaginary part of
            the DC bin, by default True
        """
        self.sample_rate = sample_rate
        self.hopsize_ms = hopsize_ms
        self.sample_rate = sample_rate
        self.hopsize_ms = hopsize_ms
        self.overlap_ratio = overlap_ratio
        self.win_oversamp = win_oversamp
        self.unpack_niquist = unpack_niquist

    def __call__(self, x: Tensor) -> Tensor:
        """
        Parameters
        ----------
        x : Tensor
            Signal of shape (B, 2*C, T, F)

        Returns
        -------
        Tensor
            ISTFT of shape (B, C, T')
        """
        x = pack_complex(x)

        if self.unpack_niquist:
            # putting back the niquist at its place
            x = F.pad(x, (0, 1))
            x[..., -1] = x[..., 0].imag + 0j
            x[..., 0] = x[..., 0].real + 0j

        batches, channels = x.shape[:2]
        x = x.flatten(0, 1)

        x = istft(
            x,
            sample_rate=self.sample_rate,
            hopsize_ms=self.hopsize_ms,
            win_len_ms=self.hopsize_ms * self.overlap_ratio,
            win_oversamp=self.win_oversamp,
        )
        x = x.unflatten(0, (batches, channels))
        return x



from typing import Iterable, Tuple
from torch import Tensor, nn
import numpy as np
import itertools
import unittest
import torch

import torch_utilities as tu
from tests.generate_test_data import get_test_data_dir


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestPyTorch(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass

    @torch.set_grad_enabled(True)
    def test_compute_gradients(self):
        keep_graph = (False, True)
        for k in keep_graph:
            x = torch.ones((3, 2), requires_grad=True)
            y = x**2
            grad = tu.compute_gradients(x, y, keep_graph=k)
            z = grad.sum()
            self.assertLess(z.item() - 2 * 6, 1e-8)
            if k:
                z.backward()
            else:
                with self.assertRaises(RuntimeError):
                    z.backward()

    def test_get_submodules(self):
        model = nn.Sequential(
            nn.Identity(),
            nn.ReLU(),
            nn.Tanh(),
            nn.Sequential(
                nn.SELU(),
                nn.Sigmoid(),
            ),
        )
        modules = tu.get_submodules(model)
        modules_types = [type(m) for m in modules]
        expected = [nn.Identity, nn.ReLU, nn.Tanh, nn.SELU, nn.Sigmoid]
        self.assertListEqual(modules_types, expected)

    def test_load_checkpoints(self):
        checkpoint_file = get_test_data_dir() / "dummy_ckpt.yml"
        checkpoints = tu.load_checkpoints(checkpoint_file)
        self.assertEqual(len(checkpoints), 10)
        [self.assertEqual(x[0], f"checkpoint_{9-i}") for i, x in enumerate(checkpoints)]
        [
            self.assertLess(np.abs(x[1] - 9 + i), 1e-12)
            for i, x in enumerate(checkpoints)
        ]

    def test_sort_checkpoints(self):
        names = ("a", "b")
        scores = (-10.2, 11)
        checkpoints = list(zip(names, scores))
        checkpoints = tu.sort_checkpoints(checkpoints)
        self.assertEqual(checkpoints[0][0], "b")
        self.assertEqual(checkpoints[1][0], "a")

    @torch.set_grad_enabled(True)
    def test_freeze_model(self):
        module = nn.Conv2d(2, 2, 1)
        self.assertTrue(all(p.requires_grad for p in module.parameters()))
        tu.freeze_model(module)
        self.assertFalse(all(p.requires_grad for p in module.parameters()))


class TestCosineScheduler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        # self.values = [(0, 1), (100, -100)]
        self.values = [
            (0, 1),
        ]
        # self.total_epochs = (10, 21)
        self.total_epochs = (10,)
        # self.iterations_per_epochs = (1, 7)
        self.iterations_per_epochs = (1,)
        # self.warmup_epochs = (0, 3)
        self.warmup_epochs = (0,)
        self.params = itertools.product(
            self.values,
            self.total_epochs,
            self.iterations_per_epochs,
            self.warmup_epochs,
        )

    def get_instance(self, params: Tuple) -> tu.CosineScheduler:
        vs, e, it, we = params
        scheduler = tu.CosineScheduler(*vs, e, it, we)
        return scheduler

    def test_compute_schedule(self) -> None:
        for p in self.params:
            vs, e, it, we = p
            a, b = vs
            with self.subTest(p=p):
                scheduler = self.get_instance(p)
                scheduling = scheduler.schedule
                warmup_steps = we * it
                # warmup
                warmup_start = a if (warmup_steps == 0) else (a / warmup_steps)
                self.assertLess(np.abs(warmup_start - scheduling[0]), 1e-12)
                # start
                self.assertLess(np.abs(scheduling[warmup_steps] - a), 1e-12)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from typing import Iterable, Tuple
from torch import Tensor
import numpy as np
import itertools
import unittest
import torch

import torch_utilities as tu
from torch_utilities.features import STFT, ISTFT


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestStftIstft(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.params = self.get_params()

    def get_params(self) -> Iterable:
        sample_rate = (16000, 32000, 48000)
        hopsize_ms = (4, 8)
        overlap_ratio = (2, 4)
        win_oversamp = (1, 2)
        pack_niquist = (False, True)

        params = itertools.product(
            sample_rate,
            hopsize_ms,
            overlap_ratio,
            win_oversamp,
            pack_niquist,
        )
        return params

    def get_input(self, params: Tuple) -> Tensor:
        sample_rate = params[0]
        x = torch.randn(1, 1, sample_rate // 4)
        return x

    def get_instances(self, p: Tuple) -> Tuple[STFT, ISTFT]:
        (
            sample_rate,
            hopsize_ms,
            overlap_ratio,
            win_oversamp,
            pack_niquist,
        ) = p
        stft = STFT(sample_rate, hopsize_ms, overlap_ratio, win_oversamp, pack_niquist)
        istft = ISTFT(
            sample_rate, hopsize_ms, overlap_ratio, win_oversamp, pack_niquist
        )
        return stft, istft

    def test_inversion(self):
        for p in self.params:
            with self.subTest(p=p):
                stft, istft = self.get_instances(p)
                x = self.get_input(p)
                y = stft(x)
                x_hat = istft(y)
                x = x[..., : x_hat.shape[-1]]
                x_hat = x_hat[..., : x.shape[-1]]
                e_max = (x - x_hat).abs().max()
                self.assertLess(e_max, 1e-5)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from torch import Tensor

from torch_utilities.common import pack_complex, phase


# export list
__all__ = ["ccmse_loss", "kld_normal_loss"]


def ccmse_loss(c: float = 0.3) -> Tensor:
    """
    Complex Compressed Mean-Squared Error loss
    | TOWARDS EFFICIENT MODELS FOR REAL-TIME DEEP NOISE SUPPRESSION
    | Authors: S. Braun et. al.

    Parameters
    ----------
    y_hat : Tensor
        Prediction of shape (B, 2, T, F)
    y_true : Tensor
        Label of shape (B, 2, T, F)
    c : float
        Compression coefficient for the spectrum,
        betweem 0 and 1
    """

    def _loss(y_hat: Tensor, y_true: Tensor) -> Tensor:
        """
        y_hat : Tensor
            Prediction of shape (B, 2, T, F)
        y_true : Tensor
            Label of shape (B, 2, T, F)
        """
        y_hat, y_true = [pack_complex(y) for y in (y_hat, y_true)]

        _f = lambda x, y, z, w: (
            ((x.abs() ** c) * z - (y.abs() ** c) * w).abs() ** 2
        ).mean()
        l1 = _f(y_hat, y_true, 1, 1)
        l2 = _f(y_hat, y_true, phase(y_hat), phase(y_true))
        loss = l1 + l2
        return loss

    return ccmse_loss


def kld_normal_loss(mu: Tensor, log_var: Tensor) -> Tensor:
    """
    Kullback-Leibler divergence w.r.t a standard normal distribution.

    Parameters
    ----------
    mu : Tensor
        Mean of the other distribution shape (B, T, F)
    log_var : Tensor
        Natural logarithm of the variance of the other distribution
        shape (B, T, F)

    Returns
    -------
    Tensor
        KLD loss between a normal distribution with mu=0 and logvar=0
        and another normal distribution
    """
    loss = (-0.5 * (1 + log_var - mu**2 - log_var.exp())).mean()
    return loss



from itertools import product
import onnxruntime as ort
import numpy as np
import unittest
import torch

from torch_utilities.metrics import DNSMOS
import torch_utilities as tu


def _setup() -> None:
    torch.manual_seed(984)
    np.random.seed(901)
    tu.set_device("auto")
    torch.set_grad_enabled(False)


class TestMetrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        pass


class TestDNSMOS(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        _setup()

    def setUp(self):
        self.dnsmos = DNSMOS()

    def test_get_inference_session(self):
        sess = self.dnsmos.get_inference_session()
        self.assertEqual(type(sess), ort.InferenceSession)

    def test_call(self):
        module = (np, torch)
        channels = (1, 2)
        length = (320000,)
        for p in product(module, channels, length):
            m, c, t = p
            with self.subTest(p=p):
                x = m.zeros((c, t))
                scores = self.dnsmos(x)
                self.assertEqual(len(scores), 3)


if __name__ == "__main__":
    unittest.main(verbosity=2)
